# Copyright 2011 Omniscale (http://omniscale.com)
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import relations with missing rings
import_partial_relations = False

# select relation builder: union or contains
relation_builder = 'contains'

# log relation that take longer than x seconds
imposm_multipolygon_report = 60

# skip relations with more rings (0 skip nothing)
imposm_multipolygon_max_ring = 0
